from functools import partial

from .. import query
from ..graph import Link, Nothing
from ..types import TypeRef, Sequence
from ..query import merge
from ..types import Any
from ..engine import Query
from ..expr.refs import RequirementsExtractor
from ..expr.core import to_expr, S
from ..expr.checker import check, graph_types, fn_types
from ..expr.compiler import ExpressionCompiler


THIS_LINK_NAME = '__link_to_this'


def _create_result_proc(engine_query, procs, options):
    def result_proc():
        sq_result = engine_query.result()
        return [[proc(this, sq_result, *opt_args)
                 for proc, opt_args in zip(procs, options)]
                for this in sq_result[THIS_LINK_NAME]]
    return result_proc


def _yield_options(query_field, graph_field):
    options = query_field.options or {}
    for option in graph_field.options:
        value = options.get(option.name, option.default)
        if value is Nothing:
            raise TypeError('Required option "{}" for {!r} was not provided'
                            .format(option.name, graph_field))
        else:
            yield value


class BoundExpr(object):

    def __init__(self, sub_graph, expr):
        self.sub_graph = sub_graph
        self.expr = expr

    def __postprocess__(self, field):
        expr, funcs = to_expr(self.expr)

        types = self.sub_graph.types.copy()
        types.update(fn_types(funcs))
        types.update((opt.name, Any) for opt in field.options)

        expr = check(expr, types)

        option_names_set = set(opt.name for opt in field.options)
        reqs = RequirementsExtractor.extract(types, expr)
        reqs = query.Node([f for f in reqs.fields
                           if f.name not in option_names_set])

        option_names = [opt.name for opt in field.options]
        code = ExpressionCompiler.compile_lambda_expr(expr, option_names)
        proc = partial(eval(compile(code, '<expr>', 'eval')),
                       {f.__def_name__: f.__def_body__ for f in funcs})
        field.func = CheckedExpr(self.sub_graph, reqs, proc)

    def __call__(self, *args, **kwargs):
        raise TypeError('Expression is not checked: {!r}'.format(self.expr))


class CheckedExpr(object):

    def __init__(self, sub_graph, reqs, proc):
        self.sub_graph = sub_graph
        self.reqs = reqs
        self.proc = proc

    @property
    def __subquery__(self):
        return self.sub_graph


class SubGraph(object):

    def __init__(self, graph, node):
        self.graph = graph
        self.node = node

        types = graph_types(graph)
        types['this'] = types[node]  # make an alias
        self.types = types

    @property
    def __subquery__(self):
        return self

    def __postprocess__(self, field):
        BoundExpr(self, getattr(S.this, field.name)).__postprocess__(field)

    def __call__(self, graph_fields, query_fields, option_values, ids,
                 queue, ctx, task_set):
        this_link = Link(THIS_LINK_NAME, Sequence[TypeRef[self.node]],
                         None, requires=None)

        reqs = merge(f.func.reqs for f in graph_fields)
        procs = [f.func.proc for f in graph_fields]

        this_req = reqs.fields_map['this'].node
        other_reqs = query.Node([r for r in reqs.fields
                                 if r.name != 'this'])

        q = Query(queue, task_set, self.graph, ctx)
        q.process_link(self.graph.root, this_link, this_req, None, ids)
        q.process_node(self.graph.root, other_reqs, None)
        return _create_result_proc(q, procs, option_values)

    def compile(self, expr):
        return BoundExpr(self, expr)

    def c(self, expr):
        return self.compile(expr)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hiku/protobuf/query.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hiku/protobuf/query.proto',
  package='hiku.protobuf.query',
  syntax='proto3',
  serialized_pb=_b('\n\x19hiku/protobuf/query.proto\x12\x13hiku.protobuf.query\"\x1f\n\x0eRepeatedString\x12\r\n\x05items\x18\x01 \x03(\t\" \n\x0fRepeatedInteger\x12\r\n\x05items\x18\x01 \x03(\x11\"\xb8\x01\n\x06Option\x12\x10\n\x06string\x18\x01 \x01(\tH\x00\x12\x11\n\x07integer\x18\x02 \x01(\x11H\x00\x12>\n\x0frepeated_string\x18\x03 \x01(\x0b\x32#.hiku.protobuf.query.RepeatedStringH\x00\x12@\n\x10repeated_integer\x18\x04 \x01(\x0b\x32$.hiku.protobuf.query.RepeatedIntegerH\x00\x42\x07\n\x05value\"\x9c\x01\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x38\n\x07options\x18\x02 \x03(\x0b\x32\'.hiku.protobuf.query.Field.OptionsEntry\x1aK\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.hiku.protobuf.query.Option:\x02\x38\x01\"\xc3\x01\n\x04Link\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\'\n\x04node\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.Node\x12\x37\n\x07options\x18\x03 \x03(\x0b\x32&.hiku.protobuf.query.Link.OptionsEntry\x1aK\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0b\x32\x1b.hiku.protobuf.query.Option:\x02\x38\x01\"g\n\x04Item\x12+\n\x05\x66ield\x18\x01 \x01(\x0b\x32\x1a.hiku.protobuf.query.FieldH\x00\x12)\n\x04link\x18\x02 \x01(\x0b\x32\x19.hiku.protobuf.query.LinkH\x00\x42\x07\n\x05value\"0\n\x04Node\x12(\n\x05items\x18\x01 \x03(\x0b\x32\x19.hiku.protobuf.query.Itemb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_REPEATEDSTRING = _descriptor.Descriptor(
  name='RepeatedString',
  full_name='hiku.protobuf.query.RepeatedString',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='hiku.protobuf.query.RepeatedString.items', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=50,
  serialized_end=81,
)


_REPEATEDINTEGER = _descriptor.Descriptor(
  name='RepeatedInteger',
  full_name='hiku.protobuf.query.RepeatedInteger',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='hiku.protobuf.query.RepeatedInteger.items', index=0,
      number=1, type=17, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=115,
)


_OPTION = _descriptor.Descriptor(
  name='Option',
  full_name='hiku.protobuf.query.Option',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string', full_name='hiku.protobuf.query.Option.string', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='integer', full_name='hiku.protobuf.query.Option.integer', index=1,
      number=2, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='repeated_string', full_name='hiku.protobuf.query.Option.repeated_string', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='repeated_integer', full_name='hiku.protobuf.query.Option.repeated_integer', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='hiku.protobuf.query.Option.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=118,
  serialized_end=302,
)


_FIELD_OPTIONSENTRY = _descriptor.Descriptor(
  name='OptionsEntry',
  full_name='hiku.protobuf.query.Field.OptionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='hiku.protobuf.query.Field.OptionsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='hiku.protobuf.query.Field.OptionsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=386,
  serialized_end=461,
)

_FIELD = _descriptor.Descriptor(
  name='Field',
  full_name='hiku.protobuf.query.Field',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hiku.protobuf.query.Field.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='options', full_name='hiku.protobuf.query.Field.options', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_FIELD_OPTIONSENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=305,
  serialized_end=461,
)


_LINK_OPTIONSENTRY = _descriptor.Descriptor(
  name='OptionsEntry',
  full_name='hiku.protobuf.query.Link.OptionsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='hiku.protobuf.query.Link.OptionsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='hiku.protobuf.query.Link.OptionsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=386,
  serialized_end=461,
)

_LINK = _descriptor.Descriptor(
  name='Link',
  full_name='hiku.protobuf.query.Link',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hiku.protobuf.query.Link.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='node', full_name='hiku.protobuf.query.Link.node', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='options', full_name='hiku.protobuf.query.Link.options', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_LINK_OPTIONSENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=464,
  serialized_end=659,
)


_ITEM = _descriptor.Descriptor(
  name='Item',
  full_name='hiku.protobuf.query.Item',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='field', full_name='hiku.protobuf.query.Item.field', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='link', full_name='hiku.protobuf.query.Item.link', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='hiku.protobuf.query.Item.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=661,
  serialized_end=764,
)


_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='hiku.protobuf.query.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='items', full_name='hiku.protobuf.query.Node.items', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=766,
  serialized_end=814,
)

_OPTION.fields_by_name['repeated_string'].message_type = _REPEATEDSTRING
_OPTION.fields_by_name['repeated_integer'].message_type = _REPEATEDINTEGER
_OPTION.oneofs_by_name['value'].fields.append(
  _OPTION.fields_by_name['string'])
_OPTION.fields_by_name['string'].containing_oneof = _OPTION.oneofs_by_name['value']
_OPTION.oneofs_by_name['value'].fields.append(
  _OPTION.fields_by_name['integer'])
_OPTION.fields_by_name['integer'].containing_oneof = _OPTION.oneofs_by_name['value']
_OPTION.oneofs_by_name['value'].fields.append(
  _OPTION.fields_by_name['repeated_string'])
_OPTION.fields_by_name['repeated_string'].containing_oneof = _OPTION.oneofs_by_name['value']
_OPTION.oneofs_by_name['value'].fields.append(
  _OPTION.fields_by_name['repeated_integer'])
_OPTION.fields_by_name['repeated_integer'].containing_oneof = _OPTION.oneofs_by_name['value']
_FIELD_OPTIONSENTRY.fields_by_name['value'].message_type = _OPTION
_FIELD_OPTIONSENTRY.containing_type = _FIELD
_FIELD.fields_by_name['options'].message_type = _FIELD_OPTIONSENTRY
_LINK_OPTIONSENTRY.fields_by_name['value'].message_type = _OPTION
_LINK_OPTIONSENTRY.containing_type = _LINK
_LINK.fields_by_name['node'].message_type = _NODE
_LINK.fields_by_name['options'].message_type = _LINK_OPTIONSENTRY
_ITEM.fields_by_name['field'].message_type = _FIELD
_ITEM.fields_by_name['link'].message_type = _LINK
_ITEM.oneofs_by_name['value'].fields.append(
  _ITEM.fields_by_name['field'])
_ITEM.fields_by_name['field'].containing_oneof = _ITEM.oneofs_by_name['value']
_ITEM.oneofs_by_name['value'].fields.append(
  _ITEM.fields_by_name['link'])
_ITEM.fields_by_name['link'].containing_oneof = _ITEM.oneofs_by_name['value']
_NODE.fields_by_name['items'].message_type = _ITEM
DESCRIPTOR.message_types_by_name['RepeatedString'] = _REPEATEDSTRING
DESCRIPTOR.message_types_by_name['RepeatedInteger'] = _REPEATEDINTEGER
DESCRIPTOR.message_types_by_name['Option'] = _OPTION
DESCRIPTOR.message_types_by_name['Field'] = _FIELD
DESCRIPTOR.message_types_by_name['Link'] = _LINK
DESCRIPTOR.message_types_by_name['Item'] = _ITEM
DESCRIPTOR.message_types_by_name['Node'] = _NODE

RepeatedString = _reflection.GeneratedProtocolMessageType('RepeatedString', (_message.Message,), dict(
  DESCRIPTOR = _REPEATEDSTRING,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.RepeatedString)
  ))
_sym_db.RegisterMessage(RepeatedString)

RepeatedInteger = _reflection.GeneratedProtocolMessageType('RepeatedInteger', (_message.Message,), dict(
  DESCRIPTOR = _REPEATEDINTEGER,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.RepeatedInteger)
  ))
_sym_db.RegisterMessage(RepeatedInteger)

Option = _reflection.GeneratedProtocolMessageType('Option', (_message.Message,), dict(
  DESCRIPTOR = _OPTION,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Option)
  ))
_sym_db.RegisterMessage(Option)

Field = _reflection.GeneratedProtocolMessageType('Field', (_message.Message,), dict(

  OptionsEntry = _reflection.GeneratedProtocolMessageType('OptionsEntry', (_message.Message,), dict(
    DESCRIPTOR = _FIELD_OPTIONSENTRY,
    __module__ = 'hiku.protobuf.query_pb2'
    # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Field.OptionsEntry)
    ))
  ,
  DESCRIPTOR = _FIELD,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Field)
  ))
_sym_db.RegisterMessage(Field)
_sym_db.RegisterMessage(Field.OptionsEntry)

Link = _reflection.GeneratedProtocolMessageType('Link', (_message.Message,), dict(

  OptionsEntry = _reflection.GeneratedProtocolMessageType('OptionsEntry', (_message.Message,), dict(
    DESCRIPTOR = _LINK_OPTIONSENTRY,
    __module__ = 'hiku.protobuf.query_pb2'
    # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Link.OptionsEntry)
    ))
  ,
  DESCRIPTOR = _LINK,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Link)
  ))
_sym_db.RegisterMessage(Link)
_sym_db.RegisterMessage(Link.OptionsEntry)

Item = _reflection.GeneratedProtocolMessageType('Item', (_message.Message,), dict(
  DESCRIPTOR = _ITEM,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Item)
  ))
_sym_db.RegisterMessage(Item)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), dict(
  DESCRIPTOR = _NODE,
  __module__ = 'hiku.protobuf.query_pb2'
  # @@protoc_insertion_point(class_scope:hiku.protobuf.query.Node)
  ))
_sym_db.RegisterMessage(Node)


_FIELD_OPTIONSENTRY.has_options = True
_FIELD_OPTIONSENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
_LINK_OPTIONSENTRY.has_options = True
_LINK_OPTIONSENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
# @@protoc_insertion_point(module_scope)



import sys

from .compat import qualname


def kw_only(fn, mapping, required, optional=None):
    d = mapping.copy()

    missing = [a for a in required if a not in d]
    if len(missing) == 1:
        raise TypeError('{}() missing 1 required keyword-only argument: {!r}'
                        .format(qualname(fn), missing[0]))
    elif missing:
        names = ', '.join(map(repr, missing[:-1]))
        raise TypeError('{}() missing {} required keyword-only arguments: {} '
                        'and {}'.format(qualname(fn), len(missing), names,
                                        repr(missing[-1])))

    result = [d.pop(key) for key in required]
    result.extend(d.pop(key, default) for key, default in (optional or ()))

    unexpected = sorted(d)
    if len(unexpected) == 1:
        name, = d
        raise TypeError('{}() got 1 unexpected keyword argument: {!r}'
                        .format(qualname(fn), name))
    elif unexpected:
        names = ', '.join(map(repr, unexpected[:-1]))
        raise TypeError('{}() got {} unexpected keyword arguments: {} and {}'
                        .format(qualname(fn), len(unexpected), names,
                                repr(unexpected[-1])))

    return result


class cached_property(object):

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value


def const(name):
    t = type(name, (object,), {})
    t.__module__ = sys._getframe(1).f_globals.get('__name__', '__main__')
    return t



from ..query import Node, Link, Field
from ..protobuf import query_pb2


_OPTION_VALUE_GETTERS = {
    'string': lambda op: op.string,
    'integer': lambda op: op.integer,
    'repeated_string': lambda op: list(op.repeated_string.items),
    'repeated_integer': lambda op: list(op.repeated_integer.items),
}


def _transform_options(pb_obj):
    options = {}
    for key, pb_option in pb_obj.options.items():
        option_type = pb_option.WhichOneof('value')
        if option_type is None:
            raise TypeError('Option value is not set: {!r}'.format(pb_option))
        options[key] = _OPTION_VALUE_GETTERS[option_type](pb_option)
    return options or None


def transform(pb_node):
    fields = []
    for i in pb_node.items:
        item_type = i.WhichOneof('value')
        if item_type == 'field':
            if not i.field.name:
                raise TypeError('Field name is empty: {!r}'.format(i))
            fields.append(Field(i.field.name, _transform_options(i.field)))
        elif item_type == 'link':
            if not i.link.name:
                raise TypeError('Link name is empty: {!r}'.format(i))
            fields.append(Link(i.link.name, transform(i.link.node),
                               _transform_options(i.link)))
        else:
            raise TypeError('Node item is empty: {!r}'.format(i))
    return Node(fields)


def read(data):
    pb_value = query_pb2.Node.FromString(data)
    return transform(pb_value)



from functools import wraps
from itertools import chain
from collections import namedtuple

from ..edn import loads
from ..query import Node, Link, Field
from ..types import Record, Callable, Any
from ..compat import text_type, string_types
from ..readers.simple import transform

from .nodes import Symbol, Tuple, List, Keyword, Dict


_Func = namedtuple('__func__', 'expr, args')


class _DotHandler(object):

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return repr(self.obj)

    def __getattr__(self, name):
        return _DotHandler(Tuple([Symbol('get'), self.obj, Symbol(name)]))


class _S(object):

    def __getattr__(self, name):
        return _DotHandler(Symbol(name))


S = _S()


def _to_expr(obj, fn_reg):
    if isinstance(obj, _DotHandler):
        return obj.obj
    elif isinstance(obj, _Func):
        fn_reg.add(obj.expr)
        return Tuple([Symbol(obj.expr.__def_name__)] +
                     [_to_expr(arg, fn_reg) for arg in obj.args])
    elif isinstance(obj, list):
        return List(_to_expr(v, fn_reg) for v in obj)
    elif isinstance(obj, dict):
        values = chain.from_iterable((Keyword(k), _to_expr(v, fn_reg))
                                     for k, v in obj.items())
        return Dict(values)
    else:
        return obj


def to_expr(obj):
    functions = set([])
    node = _to_expr(obj, functions)
    return node, tuple(functions)


def _query_to_types(obj):
    if isinstance(obj, Node):
        return Record[[(f.name, _query_to_types(f)) for f in obj.fields]]
    elif isinstance(obj, Link):
        return _query_to_types(obj.node)
    elif isinstance(obj, Field):
        return Any
    else:
        raise TypeError(type(obj))


def define(*requires, **kwargs):
    def decorator(fn):
        _name = kwargs.pop('_name', None)
        assert not kwargs, repr(kwargs)

        name = _name or '{}/{}_{}'.format(fn.__module__, fn.__name__, id(fn))

        @wraps(fn)
        def expr(*args):
            return _Func(expr, args)

        expr.__def_name__ = name
        expr.__def_body__ = fn

        if len(requires) == 1 and isinstance(requires[0], string_types):
            reqs_list = loads(text_type(requires[0]))
            expr.__def_type__ = Callable[[(_query_to_types(transform(r))
                                           if r is not None else Any)
                                          for r in reqs_list]]
        else:
            expr.__def_type__ = Callable[requires]

        return expr
    return decorator


@define(Any, Any, Any, _name='each')
def each(var, col, expr):
    pass


@define(Any, Any, Any, _name='if')
def if_(test, then, else_):
    pass


@define(Any, Any, Any, _name='if_some')
def if_some(bind, then, else_):
    pass



from ..edn import Keyword, Dict, Tuple, List
from ..query import QueryVisitor


class Exporter(QueryVisitor):

    def visit_field(self, obj):
        f = Keyword(obj.name)
        if obj.options is not None:
            f = Tuple([f, Dict((Keyword(k), v)
                               for k, v in obj.options.items())])
        return f

    def visit_link(self, obj):
        l = Keyword(obj.name)
        if obj.options is not None:
            l = Tuple([l, Dict((Keyword(k), v)
                               for k, v in obj.options.items())])
        return Dict([(l, self.visit(obj.node))])

    def visit_node(self, obj):
        return List(self.visit(f) for f in obj.fields)


def export(query):
    return Exporter().visit(query)



"""
    hiku.query
    ~~~~~~~~~~

    `Hiku` doesn't rely on any specific query language, internally it uses
    generic query representation to describe result and it could be constructed
    by parsing different suitable query languages.

    However, `Hiku` provides one built-in way to describe result -- `edn`_ data
    structure -- `simple` queries, which are very similar to the `om.next`_
    queries, which are inspired by `Datomic Pull API`_.

      - ``[:foo]`` - node fields definition (`edn` keywords in vector)
      - ``{:bar [:baz]}`` - link definition (`edn` map with keyword and vector)
      - ``(:foo {:key val})`` - field or link options definition (field name or
        link name, wrapped with `edn` list with map of options as a second
        value)

    Example:

    .. code-block:: clojure

        [:foo {:bar [:baz]}]

    This query will be read internally as:

    .. code-block:: python

        Node([Field('foo'),
              Link('bar', Node([Field('baz')]))])

    And query result will look like this:

    .. code-block:: python

        {
            'foo': 1,
            'bar': {
                'baz': 2,
            },
        }

    .. _edn: https://github.com/edn-format/edn
    .. _Datomic Pull API: http://docs.datomic.com/pull.html
    .. _om.next: https://github.com/omcljs/om/wiki/Documentation-(om.next)
"""
from itertools import chain
from collections import OrderedDict

from .utils import cached_property


def _name_repr(name, options):
    if options is None:
        return ':{}'.format(name)
    else:
        options_repr = ' '.join((':{} {!r}'.format(k, v)
                                 for k, v in options.items()))
        return '(:{} {{{}}})'.format(name, options_repr)


class Field(object):
    """Represents a field of the node

    :param name: name of the field
    :param optional options: field options -- mapping of names to values
    """
    def __init__(self, name, options=None):
        self.name = name
        self.options = options

    def __repr__(self):
        return _name_repr(self.name, self.options)

    def accept(self, visitor):
        return visitor.visit_field(self)


class Link(object):
    """Represents a link to the node

    :param name: name of the link
    :param node: collection of fields and links --
                 :py:class:`~hiku.query.Node`
    :param optional options: link options -- mapping of names to values
    """
    def __init__(self, name, node, options=None):
        self.name = name
        self.node = node
        self.options = options

    def __repr__(self):
        return '{{{} {!r}}}'.format(_name_repr(self.name, self.options),
                                    self.node)

    def accept(self, visitor):
        return visitor.visit_link(self)


class Node(object):
    """Represents collection of fields and links

    :param fields: list of :py:class:`~hiku.query.Field` and
                   :py:class:`~hiku.query.Link`
    """
    def __init__(self, fields):
        self.fields = fields

    @cached_property
    def fields_map(self):
        return OrderedDict((f.name, f) for f in self.fields)

    def __repr__(self):
        return '[{}]'.format(' '.join(map(repr, self.fields)))

    def accept(self, visitor):
        return visitor.visit_node(self)


def _merge(nodes):
    seen = set()
    to_merge = OrderedDict()
    for field in chain.from_iterable(e.fields for e in nodes):
        if field.__class__ is Link:
            to_merge.setdefault(field.name, []).append(field.node)
        else:
            if field.name not in seen:
                seen.add(field.name)
                yield field
    for name, values in to_merge.items():
        yield Link(name, Node(list(_merge(values))))


def merge(nodes):
    """Merges multiple queries into one query

    :param nodes: queries, represented as list of :py:class:`~hiku.query.Node`
    :return: merged query as one :py:class:`~hiku.query.Node`
    """
    return Node(list(_merge(nodes)))


class QueryVisitor(object):

    def visit(self, obj):
        return obj.accept(self)

    def visit_field(self, obj):
        pass

    def visit_link(self, obj):
        self.visit(obj.node)

    def visit_node(self, obj):
        for item in obj.fields:
            self.visit(item)



import collections

from itertools import repeat
from contextlib import contextmanager

from ..types import AbstractTypeVisitor
from ..query import QueryVisitor
from ..graph import GraphVisitor, Node, Field, Link, Nothing
from ..compat import text_type

from .errors import Errors


_undefined = object()


class _AssumeRecord(AbstractTypeVisitor):

    def __init__(self, _nested=False):
        self._nested = _nested

    def visit(self, obj):
        if obj is not None:
            return obj.accept(self)

    def _false(self, obj):
        pass

    visit_boolean, visit_string, visit_integer, visit_mapping, \
        visit_callable, visit_typeref = repeat(_false, 6)

    def visit_optional(self, obj):
        if not self._nested:
            return _AssumeRecord(_nested=True).visit(obj.__type__)

    def visit_sequence(self, obj):
        if not self._nested:
            return _AssumeRecord(_nested=True).visit(obj.__item_type__)

    def visit_record(self, obj):
        return list(obj.__field_types__)


class _AssumeField(GraphVisitor):

    def __init__(self, node, errors):
        self.node = node
        self.errors = errors

    def visit_field(self, obj):
        return True

    def visit_link(self, obj):
        self.errors.report('Trying to query "{}.{}" link as it was a field'
                           .format(self.node.name or 'root', obj.name))
        return False

    def visit_node(self, obj):
        assert self.node.name is None,\
            'Nested node can be only in the root node'
        self.errors.report('Trying to query "{}" node as it was a field'
                           .format(obj.name))
        return False

    def visit_root(self, obj):
        raise AssertionError('Root node is not expected here')


class _OptionError(TypeError):

    def __init__(self, description):
        self.description = description
        super(_OptionError, self).__init__(description)


class _OptionTypeError(_OptionError):

    def __init__(self, value, expected):
        description = '"{}" instead of {!r}'.format(type(value).__name__,
                                                    expected)
        super(_OptionTypeError, self).__init__(description)


class _OptionTypeValidator(object):

    def __init__(self, value):
        self._value = [value]

    @property
    def value(self):
        return self._value[-1]

    @contextmanager
    def push(self, value):
        self._value.append(value)
        try:
            yield
        finally:
            self._value.pop()

    def visit(self, type_):
        type_.accept(self)

    def visit_boolean(self, type_):
        if not isinstance(self.value, bool):
            raise _OptionTypeError(self.value, type_)

    def visit_string(self, type_):
        if not isinstance(self.value, text_type):
            raise _OptionTypeError(self.value, type_)

    def visit_integer(self, type_):
        if not isinstance(self.value, int):
            raise _OptionTypeError(self.value, type_)

    def visit_optional(self, type_):
        if self.value is not None:
            self.visit(type_.__type__)

    def visit_sequence(self, type_):
        if not isinstance(self.value, collections.Sequence):
            raise _OptionTypeError(self.value, type_)
        for item in self.value:
            with self.push(item):
                self.visit(type_.__item_type__)

    def visit_mapping(self, type_):
        if not isinstance(self.value, collections.Mapping):
            raise _OptionTypeError(self.value, type_)
        for key, value in self.value.items():
            with self.push(key):
                self.visit(type_.__key_type__)
            with self.push(value):
                self.visit(type_.__value_type__)

    def visit_record(self, type_):
        if not isinstance(self.value, collections.Mapping):
            raise _OptionTypeError(self.value, type_)

        unknown = set(self.value).difference(type_.__field_types__)
        if unknown:
            fields = ', '.join(sorted(map(repr, unknown)))
            raise _OptionError('unknown fields: {}'.format(fields))

        missing = set(type_.__field_types__).difference(self.value)
        if missing:
            fields = ', '.join(sorted(missing))
            raise _OptionError('missing fields: {}'.format(fields))

        for key, value_type in type_.__field_types__.items():
            with self.push(self.value[key]):
                self.visit(value_type)


class _ValidateOptions(GraphVisitor):

    def __init__(self, options, for_, errors):
        self.options = options
        self.for_ = for_
        self.errors = errors
        self._options = options or {}

    def visit_link(self, obj):
        super(_ValidateOptions, self).visit_link(obj)
        unknown = set(self._options).difference(obj.options_map)
        if unknown:
            node, field = self.for_
            self.errors.report('Unknown options for "{}.{}": {}'.
                               format(node, field, ', '.join(unknown)))

    visit_field = visit_link

    def visit_option(self, obj):
        value = self._options.get(obj.name, obj.default)
        if value is Nothing:
            node, field = self.for_
            self.errors.report('Required option "{}.{}:{}" is not specified'
                               .format(node, field, obj.name))
        elif obj.type is not None:
            try:
                _OptionTypeValidator(value).visit(obj.type)
            except _OptionError as err:
                node, field = self.for_
                self.errors.report('Invalid value for option "{}.{}:{}", {}'
                                   .format(node, field, obj.name,
                                           err.description))

    def visit_node(self, obj):
        assert self.options is None, 'Node can not have options'

    def visit_root(self, obj):
        raise AssertionError('Root node is not expected here')


class _RecordFieldsValidator(QueryVisitor):

    def __init__(self, field_names, errors):
        self._field_names = set(field_names)
        self._errors = errors

    def visit_field(self, obj):
        if obj.name not in self._field_names:
            self._errors.report('Unknown field name')
        elif obj.options is not None:
            self._errors.report('Options are not expected')

    def visit_link(self, obj):
        self._errors.report('Not a link')

    def visit_node(self, obj):
        raise AssertionError('Node is not expected here')


class QueryValidator(QueryVisitor):

    def __init__(self, graph):
        self.map = graph.nodes_map
        self.path = [graph.root]
        self.errors = Errors()

    def visit_field(self, obj):
        node = self.path[-1]
        field = node.fields_map.get(obj.name)
        if field is not None:
            is_field = _AssumeField(node, self.errors).visit(field)
            if is_field:
                for_ = (node.name or 'root', obj.name)
                _ValidateOptions(obj.options, for_, self.errors).visit(field)
        else:
            self.errors.report('Field "{}" is not implemented in the "{}" node'
                               .format(obj.name, node.name or 'root'))

    def visit_link(self, obj):
        node = self.path[-1]
        graph_obj = node.fields_map.get(obj.name, _undefined)
        if isinstance(graph_obj, Field):
            for_ = (node.name or 'root', obj.name)
            _ValidateOptions(obj.options, for_, self.errors).visit(graph_obj)

            field_types = _AssumeRecord().visit(graph_obj.type)
            if field_types is not None:
                fields_validator = _RecordFieldsValidator(field_types,
                                                          self.errors)
                for field in obj.node.fields:
                    fields_validator.visit(field)
            else:
                self.errors.report('Trying to query "{}.{}" simple field '
                                   'as node'
                                   .format(node.name or 'root', obj.name))

        elif isinstance(graph_obj, Link):
            linked_node = self.map[graph_obj.node]
            for_ = (node.name or 'root', obj.name)
            _ValidateOptions(obj.options, for_, self.errors).visit(graph_obj)

            self.path.append(linked_node)
            try:
                self.visit(obj.node)
            finally:
                self.path.pop()

        elif isinstance(graph_obj, Node):
            linked_node = graph_obj
            if obj.options is not None:
                self.errors.report('Options are not possible here')
            self.path.append(linked_node)
            try:
                self.visit(obj.node)
            finally:
                self.path.pop()

        elif graph_obj is _undefined:
            self.errors.report('Link "{}" is not implemented in the "{}" node'
                               .format(obj.name, node.name or 'root'))
        else:
            raise TypeError(repr(graph_obj))


def validate(graph, query):
    query_validator = QueryValidator(graph)
    query_validator.visit(query)
    return query_validator.errors.list



from __future__ import unicode_literals

from ..edn import dumps as _dumps, TaggedElement, List
from ..result import Ref, Result
from ..compat import text_type


def default(obj):
    if isinstance(obj, Ref):
        return TaggedElement('graph/ref', List((obj.node, obj.ident)))
    raise TypeError('Can not tag this object: {!r}'.format(obj))


def _transform(obj):
    if isinstance(obj, list):
        return [_transform(v) for v in obj]
    elif isinstance(obj, dict):
        assert all(isinstance(k, text_type) for k in obj.keys())
        return {k: _transform(v) for k, v in obj.items()}
    elif isinstance(obj, Result):
        return _transform(obj.root)
    else:
        return obj


def _transform_idx(idx):
    for name, value in idx.items():
        yield name, {ident: _transform(val)
                     for ident, val in value.items()}


def dumps(result, ensure_ascii=True):
    data = _transform(result)
    data.update(_transform_idx(result.index))
    return _dumps(data, default=default, ensure_ascii=ensure_ascii)



import re

from itertools import chain
from functools import partial
from collections import namedtuple, OrderedDict

from ..graph import Graph, Root, Node, Link, Option, Field, Nothing
from ..graph import GraphTransformer
from ..types import TypeRef, String, Sequence, Boolean, Optional, TypeVisitor
from ..compat import text_type, async_wrapper


LIST = namedtuple('LIST', 'of_type')
NON_NULL = namedtuple('NON_NULL', 'of_type')

FieldIdent = namedtuple('FieldIdent', 'node, name')
ArgumentIdent = namedtuple('ArgumentIdent', 'node, field, name')

NAME_RE = re.compile('^[a-zA-Z]\w*$')

ROOT_NAME = 'Root'


class TypeIdent(TypeVisitor):

    def visit_sequence(self, obj):
        return NON_NULL(LIST(self.visit(obj.__item_type__)))

    def visit_optional(self, obj):
        return self.visit(obj.__type__).of_type

    def visit_typeref(self, obj):
        return NON_NULL(obj.__type_name__)

    def visit_string(self, obj):
        return NON_NULL('String')

    def visit_integer(self, obj):
        return NON_NULL('Int')

    def visit_boolean(self, obj):
        return NON_NULL('Boolean')


def not_implemented(*args, **kwargs):
    raise NotImplementedError


def na_maybe(graph):
    return Nothing


def na_many(graph, ids=None, options=None):
    if ids is None:
        return []
    else:
        return [[] for _ in ids]


def _nodes_map(graph):
    return OrderedDict(chain(((n.name, n) for n in graph.nodes),
                             ((ROOT_NAME, graph.root),)))


def type_link(graph, options):
    name = options['name']
    if name in _nodes_map(graph):
        return name
    else:
        return Nothing


def types_link(graph):
    return [n for n in _nodes_map(graph) if NAME_RE.match(n)]


def type_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        if ident in nodes_map:
            node = nodes_map[ident]
            info = {'id': ident,
                    'kind': 'OBJECT',
                    'name': ident,
                    'description': node.description}
        elif isinstance(ident, NON_NULL):
            info = {'id': ident,
                    'kind': 'NON_NULL'}
        elif isinstance(ident, LIST):
            info = {'id': ident,
                    'kind': 'LIST'}
        else:
            assert ident in {'String', 'Int', 'Boolean'}, ident
            info = {'id': ident,
                    'name': ident,
                    'kind': 'SCALAR'}
        yield [info.get(f.name) for f in fields]


def type_fields_link(graph, ids, options):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident]
        field_idents = [FieldIdent(ident, f.name) for f in node.fields
                        if NAME_RE.match(f.name) and f.type is not None]
        if not field_idents:
            raise TypeError('Node "{}" does not contain any typed field, which '
                            'is not acceptable for GraphQL in order to define '
                            'schema type'.format(ident))
        yield field_idents


def type_of_type_link(graph, ids):
    for ident in ids:
        if isinstance(ident, (NON_NULL, LIST)):
            yield ident.of_type
        else:
            yield Nothing


def field_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        info = {'id': ident,
                'name': field.name,
                'description': field.description,
                'isDeprecated': False,
                'deprecationReason': None}
        yield [info[f.name] for f in fields]


def field_type_link(graph, ids):
    nodes_map = _nodes_map(graph)
    type_ident = TypeIdent()
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        yield type_ident.visit(field.type)


def field_args_link(graph, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        yield [ArgumentIdent(ident.node, field.name, option.name)
               for option in field.options
               if NAME_RE.match(option.name) and option.type is not None]


def input_value_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.field]
        option = field.options_map[ident.name]
        info = {'id': ident,
                'name': option.name,
                'description': None,
                'defaultValue': (None if option.default is Nothing
                                 else text_type(option.default))}
        yield [info[f.name] for f in fields]


def input_value_type_link(graph, ids):
    nodes_map = _nodes_map(graph)
    type_ident = TypeIdent()
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.field]
        option = field.options_map[ident.name]
        yield type_ident.visit(option.type)


GRAPH = Graph([
    Node('__Type', [
        Field('id', None, type_info),
        Field('kind', String, type_info),
        Field('name', String, type_info),
        Field('description', String, type_info),

        # OBJECT and INTERFACE only
        Link('fields', Sequence[TypeRef['__Field']], type_fields_link,
             requires='name',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # OBJECT only
        Link('interfaces', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # INTERFACE and UNION only
        Link('possibleTypes', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # ENUM only
        Link('enumValues', Sequence[TypeRef['__EnumValue']], na_many,
             requires='id',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # INPUT_OBJECT only
        Link('inputFields', Sequence[TypeRef['__InputValue']], na_many,
             requires='id'),

        # NON_NULL and LIST only
        Link('ofType', Optional[TypeRef['__Type']], type_of_type_link,
             requires='id'),
    ]),
    Node('__Field', [
        Field('id', None, field_info),
        Field('name', String, field_info),
        Field('description', String, field_info),

        Link('args', Sequence[TypeRef['__InputValue']], field_args_link,
             requires='id'),
        Link('type', TypeRef['__Type'], field_type_link, requires='id'),
        Field('isDeprecated', Boolean, field_info),
        Field('deprecationReason', String, field_info),
    ]),
    Node('__InputValue', [
        Field('id', None, input_value_info),
        Field('name', String, input_value_info),
        Field('description', String, input_value_info),
        Link('type', TypeRef['__Type'], input_value_type_link, requires='id'),
        Field('defaultValue', String, input_value_info),
    ]),
    Node('__Directive', [
        Field('name', String, not_implemented),
        Field('description', String, not_implemented),
        Field('locations', String, not_implemented),  # FIXME: Sequence[String]
        Link('args', Sequence[TypeRef['__InputValue']], not_implemented,
             requires=None),
    ]),
    Node('__EnumValue', [
        Field('name', String, not_implemented),
        Field('description', String, not_implemented),
        Field('isDeprecated', Boolean, not_implemented),
        Field('deprecationReason', String, not_implemented),
    ]),
    Root([
        Node('__schema', [
            Link('types', Sequence[TypeRef['__Type']], types_link,
                 requires=None),
            Link('queryType', TypeRef['__Type'], lambda graph: ROOT_NAME,
                 requires=None),
            Link('directives', Sequence[TypeRef['__Directive']], na_many,
                 requires=None),
            Link('mutationType', Optional[TypeRef['__Type']], na_maybe,
                 requires=None),
            Link('subscriptionType', Optional[TypeRef['__Type']], na_maybe,
                 requires=None),
        ]),
        Link('__type', Optional[TypeRef['__Type']], type_link, requires=None,
             options=[Option('name', String)]),
    ]),
])


class BindToGraph(GraphTransformer):

    def __init__(self, graph):
        self.graph = graph

    def visit_field(self, obj):
        field = super(BindToGraph, self).visit_field(obj)
        field.func = partial(field.func, self.graph)
        return field

    def visit_link(self, obj):
        link = super(BindToGraph, self).visit_link(obj)
        link.func = partial(link.func, self.graph)
        return link


class MakeAsync(GraphTransformer):

    def visit_field(self, obj):
        field = super(MakeAsync, self).visit_field(obj)
        field.func = async_wrapper(field.func)
        return field

    def visit_link(self, obj):
        link = super(MakeAsync, self).visit_link(obj)
        link.func = async_wrapper(link.func)
        return link


def type_name_field_func(node_name, fields, ids=None):
    return [[node_name] for _ in ids] if ids is not None else [node_name]


class AddIntrospection(GraphTransformer):

    def __init__(self, introspection_graph, type_name_field_factory):
        self.introspection_graph = introspection_graph
        self.type_name_field_factory = type_name_field_factory

    def visit_node(self, obj):
        node = super(AddIntrospection, self).visit_node(obj)
        node.fields.append(self.type_name_field_factory(obj.name))
        return node

    def visit_root(self, obj):
        root = super(AddIntrospection, self).visit_root(obj)
        root.fields.append(self.type_name_field_factory(ROOT_NAME))
        return root

    def visit_graph(self, obj):
        graph = super(AddIntrospection, self).visit_graph(obj)
        graph.items.extend(self.introspection_graph.items)
        return graph


class GraphQLIntrospection(GraphTransformer):

    def __type_name__(self, node_name):
        return Field('__typename', String,
                     partial(type_name_field_func, node_name))

    def __introspection_graph__(self, graph):
        return BindToGraph(graph).visit(GRAPH)

    def visit_node(self, obj):
        node = super(GraphQLIntrospection, self).visit_node(obj)
        node.fields.append(self.__type_name__(obj.name))
        return node

    def visit_root(self, obj):
        root = super(GraphQLIntrospection, self).visit_root(obj)
        root.fields.append(self.__type_name__(ROOT_NAME))
        return root

    def visit_graph(self, obj):
        introspection_graph = self.__introspection_graph__(obj)
        items = [self.visit(node) for node in obj.items]
        items.extend(introspection_graph.items)
        return Graph(items)


class AsyncGraphQLIntrospection(GraphQLIntrospection):

    def __type_name__(self, node_name):
        return Field('__typename', String,
                     async_wrapper(partial(type_name_field_func, node_name)))

    def __introspection_graph__(self, graph):
        introspection_graph = BindToGraph(graph).visit(GRAPH)
        introspection_graph = MakeAsync().visit(introspection_graph)
        return introspection_graph



from setuptools import setup, find_packages

setup(
    name='Hiku',
    version='0.4.0rc2',
    description='Library to implement Graph APIs',
    author='Vladimir Magamedov',
    author_email='vladimir@magamedov.com',
    url='https://github.com/vmagamedov/hiku',
    packages=find_packages(),
    include_package_data=True,
    license='BSD',
    install_requires=[],
)



import json
import string
import pkgutil
import traceback

from ..result import denormalize
from ..typedef.kinko import dumps as dumps_typedef
from ..readers.simple import read
from ..validate.query import validate


ERROR_CODES = {
    400: (
        'Bad Request',
        ('The browser (or proxy) sent a request that this server could '
         'not understand.'),
    ),
    404: (
        'Not Found',
        ('The requested URL was not found on the server.  '
         'If you entered the URL manually please check your spelling and '
         'try again.'),
    ),
    405: (
        'Method Not Allowed',
        'The method is not allowed for the requested URL.',
    ),
}


def _decode(b, charset='utf-8'):
    return b.decode(charset)


def _encode(s, charset='utf-8'):
    return s.encode(charset)


class ConsoleApplication(object):
    _urls = {
        'index_url': '/',
        'docs_url': '/docs',
        'js_url': '/console.js',
    }

    def __init__(self, root, engine, ctx=None, debug=False):
        self.root = root
        self.engine = engine
        self.ctx = ctx
        self.debug = debug
        self._console_html = string.Template(_decode(
            pkgutil.get_data('hiku.console', 'assets/console.html')
        ))
        self._docs_content = dumps_typedef(root)

    def __call__(self, environ, start_response):
        path_info = environ['PATH_INFO'] or '/'

        if path_info == self._urls['index_url']:
            if environ['REQUEST_METHOD'] == 'GET':
                return self._index_get(environ, start_response)
            elif environ['REQUEST_METHOD'] == 'POST':
                return self._index_post(environ, start_response)
            else:
                return self._error(405, start_response)

        elif path_info == self._urls['docs_url']:
            if environ['REQUEST_METHOD'] == 'GET':
                return self._docs_get(environ, start_response)
            else:
                return self._error(405, start_response)

        elif path_info == self._urls['js_url']:
            if environ['REQUEST_METHOD'] == 'GET':
                return self._static_get(environ, start_response)
            else:
                return self._error(405, start_response)

        else:
            return self._error(404, start_response)

    def _urls_map(self, environ):
        script_name = environ.get('SCRIPT_NAME', '').rstrip('/')
        return {name: '{}{}'.format(script_name, path)
                for name, path in self._urls.items()}

    def _error(self, code, start_response, message=None):
        description, standard_message = ERROR_CODES[code]
        message = message or standard_message
        content = _encode(message + '\n')
        start_response('{} {}'.format(code, description), [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(content))),
        ])
        return [content]

    def _index_get(self, environ, start_response):
        content = _encode(
            self._console_html.safe_substitute(
                **self._urls_map(environ)
            )
        )
        start_response('200 OK', [
            ('Content-Type', 'text/html'),
            ('Content-Length', str(len(content))),
        ])
        return [content]

    def _index_post(self, environ, start_response):
        limit = max(0, int(environ.get('CONTENT_LENGTH') or 0))
        if limit > 2 ** 20:  # 1MB
            return self._error(400, start_response, 'Payload is too big')

        pattern = environ['wsgi.input'].read(limit)
        try:
            # TODO: implement query validation
            query = read(_decode(pattern))

            errors = validate(self.root, query)
            if errors:
                result = {'errors': errors}
                status = '400 Bad Request'
            else:
                result = self.engine.execute(self.root, query, ctx=self.ctx)
                result = denormalize(self.root, result, query)
                status = '200 OK'
        except Exception:
            tb = traceback.format_exc() if self.debug else None
            result = {'traceback': tb}
            status = '500 Internal Server Error'
        result_data = _encode(json.dumps(result))
        start_response(status, [
            ('Content-Type', 'application/json'),
            ('Content-Length', str(len(result_data))),
        ])
        return [result_data]

    def _docs_get(self, environ, start_response):
        content = _encode(self._docs_content)
        start_response('200 OK', [
            ('Content-Length', str(len(content))),
        ])
        return [content]

    def _static_get(self, environ, start_response):
        content = pkgutil.get_data('hiku.console', 'assets/console.js')
        start_response('200 OK', [
            ('Content-Type', 'text/javascript; charset=UTF-8'),
            ('Content-Length', str(len(content))),
        ])
        return [content]
